#!/usr/bin/python
# The Utility is based on the http://bk-mg.narod.ru/ project

__author__ = 'Igor Maznitsa'

import wave
import getopt
import sys
import os
import struct

SND_PARTS = [
    bytearray([0x80, 0xbf, 0xbf, 0x80, 0x40, 0x40, 0x80, 0xbf, 0xbf, 0x80, 0x40, 0x40]),
    bytearray([0x80, 0xa0, 0xb7, 0xc0, 0xb7, 0xa0, 0x80, 0x5f, 0x48, 0x3f, 0x48, 0x5f, 0x80, 0xb7, 0xb7, 0x80, 0x48, 0x48]),
    bytearray([0x80, 0xbf, 0xbf, 0x80, 0x40, 0x40, 0x80, 0xbf, 0xbf, 0x80, 0x40, 0x40, 0x80, 0xbf, 0xbf, 0x80, 0x40, 0x40, 0x80, 0xbf, 0xbf, 0x80, 0x40, 0x40, 0x80, 0xbf, 0xbf, 0x80, 0x40, 0x40, 0x80, 0xbf, 0xbf, 0x80, 0x40, 0x40, 0x80, 0xbf, 0xbf, 0x80, 0x40, 0x40, 0x80, 0xbf, 0xbf, 0x80, 0x40, 0x40]),
    bytearray([0x80, 0x90, 0x9d, 0xa4, 0xa6, 0xa9, 0xa9, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0x80, 0x6f, 0x62, 0x5b, 0x57, 0x56, 0x55, 0x55, 0x55, 0x55, 0x55, 0x6e, 0x80, 0x9a, 0xa2, 0xa6, 0xa7, 0xa9, 0x80, 0x6f, 0x63, 0x5c, 0x59, 0x59, 0x80, 0xb7, 0xb7, 0x80, 0x48, 0x48]),
    bytearray([0x80, 0x8f, 0xa8, 0xb5, 0xbc, 0xbf, 0xc1, 0xc1, 0xc2, 0xc2, 0xc2, 0xc2, 0xc2, 0xc1, 0xc1, 0xc1, 0xc1, 0xc1, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xbf, 0xbf, 0xbe, 0xbe, 0xbe, 0xbd, 0xbc, 0xb2, 0x80, 0x59, 0x48, 0x40, 0x3b, 0x39, 0x38, 0x37, 0x37, 0x37, 0x37, 0x37, 0x38, 0x38, 0x38, 0x38, 0x39, 0x39, 0x39, 0x39, 0x39, 0x3a, 0x3a, 0x3a, 0x3a, 0x3a, 0x3b, 0x3b, 0x3b, 0x3b, 0x3c, 0x3e, 0x7a])
]

SIGNAL_RESET = 0
SIGNAL_SET = 1
SIGNAL_START_SYNCHRO = 2
SIGNAL_SYNCHRO = 3
SIGNAL_END_MARKER = 4


def amplify():
    """Extends audio diapason in the WAV signals to 1.0 instead of 0.5"""
    min_l = 256
    max_l = 0
    for arr in SND_PARTS:
        for b in arr:
            if b < min_l:
                min_l = b
            if b > max_l:
                max_l = b

    max_l -= 128
    min_l -= 128

    c_max = 128.0 / max_l
    c_min = -127.0 / min_l

    coeff = min(c_max, c_min)

    for arr in SND_PARTS:
        for i in range(len(arr)):
            arr[i] = max(0, min(255, int(round((arr[i] - 128) * coeff)) + 128))


def play(i, arr, num=1):
    """Save wav sequence for num-times into the result"""
    for c in range(num):
        for x in SND_PARTS[i]:
            arr.append(x)


def byte2arr(b, arr):
    """Append wav representation of a sound to array"""
    for i in range(0, 8):
        for a in SND_PARTS[(b >> i) & 1]:
            arr.append(a)


def txt2arr(txt, arr):
    """Write text chars as bytes into sound array (with 16 char restrictions)"""
    ws = 0
    if len(txt) < 16:
        ws = 16 - len(txt)
    if len(txt) > 16:
        txt = txt[0:16]
    for c in txt:
        byte2arr(ord(c), arr)

    while ws != 0:
        byte2arr(ord(' '), arr)
        ws -= 1


def short2arr(num, arr):
    """Write a short (0...FFFF) value into sound array)"""
    byte2arr(num & 0xFF, arr)
    byte2arr((num >> 8) & 0xFF, arr)


def init_wav(mode, filename, datalen):
    r = wave.open(filename, 'w')
    r.setnchannels(1)
    r.setsampwidth(1)
    if mode:
        r.setframerate(22050)
    else:
        r.setframerate(11025)
    r.setnframes(datalen)
    r.setcomptype('NONE', 'PCM')
    return r


def calc_crc(data):
    a = 0
    for i in data:
        a += i
        if a > 0xFFFF:
            a -= 0xFFFF
    return a


def make_bk_wav(tmode, filename, start, name, data, data_length):
    sndbuff = []

    play(SIGNAL_START_SYNCHRO, sndbuff, 512)
    play(SIGNAL_SYNCHRO, sndbuff)

    play(SIGNAL_START_SYNCHRO, sndbuff)
    play(SIGNAL_SYNCHRO, sndbuff)

    short2arr(start, sndbuff)
    short2arr(data_length, sndbuff)
    txt2arr(name, sndbuff)

    play(SIGNAL_START_SYNCHRO, sndbuff)
    play(SIGNAL_SYNCHRO, sndbuff)

    for i in range(data_length):
        byte2arr(data[i], sndbuff)

    tap_check_code = calc_crc(data)
    short2arr(tap_check_code, sndbuff)

    play(SIGNAL_END_MARKER, sndbuff)
    play(SIGNAL_START_SYNCHRO, sndbuff, 64 if tmode else 32)
    play(SIGNAL_SYNCHRO, sndbuff)

    wav_file = init_wav(tmode, filename, len(sndbuff))
    try:
        wav_file.writeframes(bytearray(sndbuff))
    finally:
        wav_file.close()
    return tap_check_code


def read_short(f):
    v = struct.unpack('BB', f.read(2))
    return (v[1] << 8) | v[0]


def read_bin(binfile):
    whole_size = os.path.getsize(binfile)

    if whole_size < 4:
        print("Wrong BIN file format detected, its size is less than 4")
        exit(1)

    f = open(binfile, 'rb')
    with open(binfile, 'rb') as f:
        saddr = read_short(f)
        slen = read_short(f)
        if slen > whole_size - 4:
            print("Wrong data size length in BIN file %d" % (slen))
            exit(1)
        arr = bytearray(f.read(slen))

        return saddr, slen, arr,


def header():
    print("""
    BK BIN2WAV Utility
-----------------------------
    The Utility converts a BIN file (the BK-0010 emulator format) into WAV version
    Author: Igor Maznitsa (http://www.igormaznitsa.com), Tallinn, Estonia, 2013

    *The Utility is a freeware and based on the http://bk-mg.narod.ru/ project*
    """)


def usage():
    print("""    Usage:

    bkbin2wav -i <binfile> [-a] [-o <wavfile>] [-n <name>] [-s addr] [-t]

    Command line options:
        -h          Print help
        -a          Amplify the audio signal in the result WAV file
        -i <file>   The BIN file to be converted
        -o <file>   The Result WAV file (by default the BIN file name with WAV extension)
        -n <name>   The Name of the file in the TAP header (must be less or equals 16 chars)
        -s <addr>   The Start address for the TAP header (by default the start address from the BIN will be used)
        -t          Use the double frequency "turbo" mode
    """)


header()

opts, args = getopt.getopt(sys.argv[1:], "athi:o:n:s:",
                           ['amplify', 'turbo', 'help', 'input', 'output', 'name', 'start'])

infile = ''
outfile = ''
addr = -1
name = ''
turbo = False
ampflag = False

for o, a in opts:
    if o in ('-h', '--help', '-?'):
        usage()
        exit(1)
    elif o == '-i':
        infile = a
    elif o == '-s':
        addr = int(a) & 0xFFFF
    elif o == '-a':
        ampflag = True
        amplify()
    elif o == '-o':
        outfile = a
    elif o == '-n':
        name = a
    elif o == '-t':
        turbo = True
    else:
        usage()
        exit(1)

if infile == '':
    print("The Input BIN file is undefied")
    exit(1)

if not os.path.isfile(infile):
    print("Can't find the BIN file %s" % (infile))
    exit(1)

if outfile == '':
    outfile = os.path.dirname(infile) + os.sep + os.path.basename(infile) + '.wav'

if name == '':
    name = os.path.splitext(os.path.basename(infile))[0].upper()[:16]

bin_start, bin_length, bin_data = read_bin(infile)

if addr == -1:
    addr = bin_start

print("""
   Input file : %s
  Output file : %s
         Name : %s
Start address : %d (%s)
   Turbo mode : %s
    Amplifier : %s
""" % (infile, outfile, name, addr, oct(addr), ('ON' if turbo else 'OFF'), ('ON' if ampflag else 'OFF')))

print("""    Fields of the BIN file header:
       Start  : %d (%s)
       Length : %d (%s)
""" % (bin_start, oct(bin_start), bin_length, oct(bin_length)))

if addr != bin_start:
    print("Warning! The Start address has been changed %d(%s)->%d(%s)" % (bin_start, oct(bin_start), addr, oct(addr)))

crc_code = make_bk_wav(turbo, outfile, addr, name, bin_data, bin_length)

print("The WAV file has been saved successfully as %s the control sum is %s" % (outfile, hex(crc_code)))